@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import os
 import time
+import typing
 import asyncio
 import logging
 import itertools
@@ -12,7 +13,9 @@ from collections import Counter, defaultdict
 import zigpy.state
 import async_timeout
 import zigpy.zdo.types as zdo_t
+from zigpy.exceptions import NetworkNotFormed
 
+import zigpy_znp
 import zigpy_znp.const as const
 import zigpy_znp.types as t
 import zigpy_znp.config as conf
@@ -30,16 +33,22 @@ from zigpy_znp.frames import GeneralFrame
 from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse
 from zigpy_znp.types.nvids import ExNvIds, OsalNvIds
 
+if typing.TYPE_CHECKING:
+    import typing_extensions
+
 LOGGER = logging.getLogger(__name__)
 
 
 # All of these are in seconds
+STARTUP_TIMEOUT = 15
 AFTER_BOOTLOADER_SKIP_BYTE_DELAY = 2.5
 NETWORK_COMMISSIONING_TIMEOUT = 30
 BOOTLOADER_PIN_TOGGLE_DELAY = 0.15
 CONNECT_PING_TIMEOUT = 0.50
 CONNECT_PROBE_TIMEOUT = 10
 
+NVRAM_MIGRATION_ID = 1
+
 
 class ZNP:
     def __init__(self, config: conf.ConfigType):
@@ -50,11 +59,11 @@ class ZNP:
         self._listeners = defaultdict(list)
         self._sync_request_lock = asyncio.Lock()
 
-        self.capabilities = None
-        self.version = None
+        self.capabilities = None  # type: int
+        self.version = None  # type: float
 
         self.nvram = NVRAMHelper(self)
-        self.network_info: zigpy.state.NetworkInformation = None
+        self.network_info: zigpy.state.NetworkInfo = None
         self.node_info: zigpy.state.NodeInfo = None
 
     def set_application(self, app):
@@ -111,17 +120,31 @@ class ZNP:
                 )
 
         if not is_on_network:
-            raise ValueError("Device is not a part of a network")
+            raise NetworkNotFormed("Device is not a part of a network")
+
+        ieee = await self.nvram.osal_read(OsalNvIds.EXTADDR, item_type=t.EUI64)
+        logical_type = await self.nvram.osal_read(
+            OsalNvIds.LOGICAL_TYPE, item_type=t.DeviceLogicalType
+        )
+
+        node_info = zigpy.state.NodeInfo(
+            ieee=ieee,
+            nwk=nib.nwkDevAddress,
+            logical_type=zdo_t.LogicalType(logical_type),
+        )
 
         key_desc = await self.nvram.osal_read(
             OsalNvIds.NWK_ACTIVE_KEY_INFO, item_type=t.NwkKeyDesc
         )
 
-        tc_frame_counter = await security.read_tc_frame_counter(
+        nwk_frame_counter = await security.read_nwk_frame_counter(
             self, ext_pan_id=nib.extendedPANID
         )
 
-        network_info = zigpy.state.NetworkInformation(
+        version = await self.request(c.SYS.Version.Req())
+
+        network_info = zigpy.state.NetworkInfo(
+            source=f"zigpy-znp@{zigpy_znp.__version__}",
             extended_pan_id=nib.extendedPANID,
             pan_id=nib.nwkPanId,
             nwk_update_id=nib.nwkUpdateId,
@@ -131,15 +154,22 @@ class ZNP:
             network_key=zigpy.state.Key(
                 key=key_desc.Key,
                 seq=key_desc.KeySeqNum,
-                tx_counter=tc_frame_counter,
-                rx_counter=None,
-                partner_ieee=None,
+                tx_counter=nwk_frame_counter,
+                rx_counter=0,
+                partner_ieee=node_info.ieee,
+            ),
+            tc_link_key=zigpy.state.Key(
+                key=const.DEFAULT_TC_LINK_KEY,
+                seq=0,
+                tx_counter=0,
+                rx_counter=0,
+                partner_ieee=node_info.ieee,
             ),
-            tc_link_key=None,
             children=[],
             nwk_addresses={},
             key_table=[],
-            stack_specific=None,
+            stack_specific={},
+            metadata={"zstack": version.as_dict()},
         )
 
         tclk_seed = None
@@ -183,31 +213,105 @@ class ZNP:
                     )
 
                 if dev.is_child:
-                    network_info.children.append(dev.node_info)
-                elif dev.node_info.nwk is not None:
+                    network_info.children.append(dev.node_info.ieee)
+
+                if dev.node_info.nwk is not None:
                     network_info.nwk_addresses[dev.node_info.ieee] = dev.node_info.nwk
 
                 if dev.key is not None:
                     network_info.key_table.append(dev.key)
 
-        ieee = await self.nvram.osal_read(OsalNvIds.EXTADDR, item_type=t.EUI64)
-        logical_type = await self.nvram.osal_read(
-            OsalNvIds.LOGICAL_TYPE, item_type=t.DeviceLogicalType
-        )
-
-        node_info = zigpy.state.NodeInfo(
-            ieee=ieee,
-            nwk=nib.nwkDevAddress,
-            logical_type=zdo_t.LogicalType(logical_type),
-        )
-
         self.network_info = network_info
         self.node_info = node_info
 
+    async def start_network(self):
+        # Both startup sequences end with the same callback
+        started_as_coordinator = self.wait_for_response(
+            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
+        )
+
+        # Handle the startup progress messages
+        async with self.capture_responses(
+            [
+                c.ZDO.StateChangeInd.Callback(
+                    State=t.DeviceState.StartingAsCoordinator
+                ),
+                c.AppConfig.BDBCommissioningNotification.Callback(
+                    partial=True,
+                    Status=c.app_config.BDBCommissioningStatus.InProgress,
+                ),
+            ]
+        ):
+            try:
+                if self.version > 1.2:
+                    # Z-Stack 3 uses the BDB subsystem
+                    commissioning_rsp = await self.request_callback_rsp(
+                        request=c.AppConfig.BDBStartCommissioning.Req(
+                            Mode=c.app_config.BDBCommissioningMode.NwkFormation
+                        ),
+                        RspStatus=t.Status.SUCCESS,
+                        callback=c.AppConfig.BDBCommissioningNotification.Callback(
+                            partial=True,
+                            RemainingModes=c.app_config.BDBCommissioningMode.NONE,
+                        ),
+                        timeout=NETWORK_COMMISSIONING_TIMEOUT,
+                    )
+
+                    # This is the correct startup sequence according to the forums,
+                    # including the formation failure error.  Success is only returned
+                    # when the network is first formed.
+                    if commissioning_rsp.Status not in (
+                        c.app_config.BDBCommissioningStatus.FormationFailure,
+                        c.app_config.BDBCommissioningStatus.Success,
+                    ):
+                        raise RuntimeError(
+                            f"Network formation failed: {commissioning_rsp}"
+                        )
+                else:
+                    # In Z-Stack 1.2.2, StartupFromApp actually does what it says
+                    rsp = await self.request(c.ZDO.StartupFromApp.Req(StartDelay=100))
+
+                    if rsp.State not in (
+                        c.zdo.StartupState.NewNetworkState,
+                        c.zdo.StartupState.RestoredNetworkState,
+                    ):
+                        raise InvalidCommandResponse(
+                            f"Invalid startup response state: {rsp.State}", rsp
+                        )
+
+                # Both versions still end with this callback
+                async with async_timeout.timeout(STARTUP_TIMEOUT):
+                    await started_as_coordinator
+            except asyncio.TimeoutError as e:
+                raise RuntimeError(
+                    "Network formation refused, RF environment is likely too noisy."
+                    " Temporarily unscrew the antenna or shield the coordinator"
+                    " with metal until a network is formed."
+                ) from e
+
+        LOGGER.debug("Waiting for NIB to stabilize")
+
+        # Even though the device says it is "ready" at this point, it takes a few more
+        # seconds for `_NIB.nwkState` to switch from `NwkState.NWK_INIT`. There does
+        # not appear to be any user-facing MT command to read this information.
+        while True:
+            try:
+                nib = await self.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
+            except KeyError:
+                pass
+            else:
+                LOGGER.debug("Current NIB is %s", nib)
+
+                # Usually this works after the first attempt
+                if nib.nwkLogicalChannel != 0 and nib.nwkPanId != 0xFFFE:
+                    break
+
+            await asyncio.sleep(1)
+
     async def write_network_info(
         self,
         *,
-        network_info: zigpy.state.NetworkInformation,
+        network_info: zigpy.state.NetworkInfo,
         node_info: zigpy.state.NodeInfo,
     ) -> None:
         """
@@ -257,82 +361,8 @@ class ZNP:
                 RspStatus=t.Status.SUCCESS,
             )
 
-        # Both startup sequences end with the same callback
-        started_as_coordinator = self.wait_for_response(
-            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
-        )
-
         LOGGER.debug("Forming temporary network")
-
-        # Handle the startup progress messages
-        async with self.capture_responses(
-            [
-                c.ZDO.StateChangeInd.Callback(
-                    State=t.DeviceState.StartingAsCoordinator
-                ),
-                c.AppConfig.BDBCommissioningNotification.Callback(
-                    partial=True,
-                    Status=c.app_config.BDBCommissioningStatus.InProgress,
-                ),
-            ]
-        ):
-            try:
-                if self.version > 1.2:
-                    # Z-Stack 3 uses the BDB subsystem
-                    commissioning_rsp = await self.request_callback_rsp(
-                        request=c.AppConfig.BDBStartCommissioning.Req(
-                            Mode=c.app_config.BDBCommissioningMode.NwkFormation
-                        ),
-                        RspStatus=t.Status.SUCCESS,
-                        callback=c.AppConfig.BDBCommissioningNotification.Callback(
-                            partial=True,
-                            RemainingModes=c.app_config.BDBCommissioningMode.NONE,
-                        ),
-                        timeout=NETWORK_COMMISSIONING_TIMEOUT,
-                    )
-
-                    if (
-                        commissioning_rsp.Status
-                        != c.app_config.BDBCommissioningStatus.Success
-                    ):
-                        raise RuntimeError(
-                            f"Network formation failed: {commissioning_rsp}"
-                        )
-                else:
-                    # In Z-Stack 1.2.2, StartupFromApp actually does what it says
-                    await self.request(
-                        c.ZDO.StartupFromApp.Req(StartDelay=100),
-                        RspState=c.zdo.StartupState.NewNetworkState,
-                    )
-
-                # Both versions still end with this callback
-                await started_as_coordinator
-            except asyncio.TimeoutError as e:
-                raise RuntimeError(
-                    "Network formation refused, RF environment is likely too noisy."
-                    " Temporarily unscrew the antenna or shield the coordinator"
-                    " with metal until a network is formed."
-                ) from e
-
-        LOGGER.debug("Waiting for NIB to stabilize")
-
-        # Even though the device says it is "ready" at this point, it takes a few more
-        # seconds for `_NIB.nwkState` to switch from `NwkState.NWK_INIT`. There does
-        # not appear to be any user-facing MT command to read this information.
-        while True:
-            try:
-                nib = await self.nvram.osal_read(OsalNvIds.NIB, item_type=t.NIB)
-            except KeyError:
-                pass
-            else:
-                LOGGER.debug("Current NIB is %s", nib)
-
-                # Usually this works after the first attempt
-                if nib.nwkLogicalChannel != 0 and nib.nwkPanId != 0xFFFE:
-                    break
-
-            await asyncio.sleep(1)
-
+        await self.start_network()
         await self.reset()
 
         LOGGER.debug("Writing actual network settings")
@@ -365,33 +395,52 @@ class ZNP:
             OsalNvIds.EXTENDED_PAN_ID: network_info.extended_pan_id,
             OsalNvIds.PRECFGKEY: key_info.Active.Key,
             OsalNvIds.CHANLIST: network_info.channel_mask,
-            OsalNvIds.EXTADDR: node_info.ieee,
+            # If the EXTADDR entry is deleted, Z-Stack resets it to the hardware address
+            OsalNvIds.EXTADDR: (
+                None if node_info.ieee == t.EUI64.UNKNOWN else node_info.ieee
+            ),
             OsalNvIds.LOGICAL_TYPE: t.DeviceLogicalType(node_info.logical_type),
             OsalNvIds.NWK_ACTIVE_KEY_INFO: key_info.Active,
-            OsalNvIds.NWK_ALTERN_KEY_INFO: const.EMPTY_KEY,
+            OsalNvIds.NWK_ALTERN_KEY_INFO: key_info.Active,
         }
 
         tclk_seed = None
 
-        if self.version > 1.2:
+        if self.version == 1.2:
+            # TCLK_SEED is TCLK_TABLE_START in Z-Stack 1
+            nvram[OsalNvIds.TCLK_SEED] = t.TCLinkKey(
+                ExtAddr=t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),  # global
+                Key=network_info.tc_link_key.key,
+                TxFrameCounter=0,
+                RxFrameCounter=0,
+            )
+        else:
+            if network_info.tc_link_key.key != const.DEFAULT_TC_LINK_KEY:
+                LOGGER.warning(
+                    "TC link key is configured at build time in Z-Stack 3 and cannot be"
+                    " changed at runtime: %s",
+                    network_info.tc_link_key.key,
+                )
+
             if (
                 network_info.stack_specific is not None
                 and network_info.stack_specific.get("zstack", {}).get("tclk_seed")
             ):
-                tclk_seed, _ = t.KeyData.deserialize(
+                tclk_seed = t.KeyData(
                     bytes.fromhex(network_info.stack_specific["zstack"]["tclk_seed"])
                 )
             else:
                 tclk_seed = t.KeyData(os.urandom(16))
 
             nvram[OsalNvIds.TCLK_SEED] = tclk_seed
-        else:
-            nvram[OsalNvIds.TCLK_SEED] = const.DEFAULT_TC_LINK_KEY
 
         for key, value in nvram.items():
-            await self.nvram.osal_write(key, value, create=True)
+            if value is None:
+                await self.nvram.osal_delete(key)
+            else:
+                await self.nvram.osal_write(key, value, create=True)
 
-        await security.write_tc_frame_counter(
+        await security.write_nwk_frame_counter(
             self,
             network_info.network_key.tx_counter,
             ext_pan_id=network_info.extended_pan_id,
@@ -399,10 +448,12 @@ class ZNP:
 
         devices = {}
 
-        for child in network_info.children or []:
-            devices[child.ieee] = security.StoredDevice(
-                node_info=dataclasses.replace(
-                    child, nwk=0xFFFE if child.nwk is None else child.nwk
+        for ieee in network_info.children or []:
+            devices[ieee] = security.StoredDevice(
+                node_info=zigpy.state.NodeInfo(
+                    nwk=network_info.nwk_addresses.get(ieee, 0xFFFE),
+                    ieee=ieee,
+                    logical_type=zdo_t.LogicalType.EndDevice,
                 ),
                 key=None,
                 is_child=True,
@@ -449,6 +500,11 @@ class ZNP:
             counter_increment=0,
         )
 
+        # Prevent an unnecessary NVRAM migration from running
+        await self.nvram.osal_write(
+            OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, t.uint8_t(NVRAM_MIGRATION_ID), create=True
+        )
+
         if self.version == 1.2:
             await self.nvram.osal_write(
                 OsalNvIds.HAS_CONFIGURED_ZSTACK1,
@@ -462,18 +518,84 @@ class ZNP:
                 create=True,
             )
 
+        # Reset after writing network settings to allow Z-Stack to recreate NVRAM items
+        # that were intentionally deleted.
+        await self.reset()
+
         LOGGER.debug("Done!")
 
-    async def reset(self) -> None:
+    async def migrate_nvram(self) -> bool:
+        """
+        Migrates NVRAM entries using the `ZIGPY_ZNP_MIGRATION_ID` NVRAM item.
+        Returns `True` if a migration was performed, `False` otherwise.
+        """
+
+        from zigpy_znp.znp import security
+
+        try:
+            migration_id = await self.nvram.osal_read(
+                OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, item_type=t.uint8_t
+            )
+        except KeyError:
+            migration_id = 0
+
+        initial_migration_id = migration_id
+
+        # Migration 1: empty `ADDRMGR` entries are version-dependent and were improperly
+        #              written for CC253x devices.
+        #
+        #              This migration is stateless and can safely be run more than once:
+        #              the only downside is that startup times increase by 10s on newer
+        #              coordinators, which is why the migration ID is persisted.
+        if migration_id < 1:
+            try:
+                entries = await security.read_addr_manager_entries(self)
+            except KeyError:
+                pass
+            else:
+                fixed_entries = []
+
+                for entry in entries:
+                    if entry.extAddr != t.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"):
+                        fixed_entries.append(entry)
+                    elif self.version == 3.30:
+                        fixed_entries.append(const.EMPTY_ADDR_MGR_ENTRY_ZSTACK3)
+                    else:
+                        fixed_entries.append(const.EMPTY_ADDR_MGR_ENTRY_ZSTACK1)
+
+                if entries != fixed_entries:
+                    LOGGER.warning(
+                        "Repairing %d invalid empty address manager entries (total %d)",
+                        sum(i != j for i, j in zip(entries, fixed_entries)),
+                        len(entries),
+                    )
+                    await security.write_addr_manager_entries(self, fixed_entries)
+
+            migration_id = 1
+
+        if initial_migration_id == migration_id:
+            return False
+
+        await self.nvram.osal_write(
+            OsalNvIds.ZIGPY_ZNP_MIGRATION_ID, t.uint8_t(migration_id), create=True
+        )
+        await self.reset()
+
+        return True
+
+    async def reset(self, *, wait_for_reset: bool = True) -> None:
         """
         Performs a soft reset within Z-Stack.
         A hard reset resets the serial port, causing the device to disconnect.
         """
 
-        await self.request_callback_rsp(
-            request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
-            callback=c.SYS.ResetInd.Callback(partial=True),
-        )
+        if wait_for_reset:
+            await self.request_callback_rsp(
+                request=c.SYS.ResetReq.Req(Type=t.ResetType.Soft),
+                callback=c.SYS.ResetInd.Callback(partial=True),
+            )
+        else:
+            await self.request(c.SYS.ResetReq.Req(Type=t.ResetType.Soft))
 
     @property
     def _port_path(self) -> str:
@@ -542,7 +664,7 @@ class ZNP:
 
                 try:
                     async with async_timeout.timeout(CONNECT_PING_TIMEOUT):
-                        result = await ping_task
+                        result = await ping_task  # type:ignore[misc]
                 except asyncio.TimeoutError:
                     ping_task.cancel()
 
@@ -587,7 +709,6 @@ class ZNP:
         """
         Called by the UART object when a connection has been made.
         """
-        pass
 
     def connection_lost(self, exc) -> None:
         """
@@ -609,7 +730,7 @@ class ZNP:
 
         self._app = None
 
-        for header, listeners in self._listeners.items():
+        for _header, listeners in self._listeners.items():
             for listener in listeners:
                 listener.cancel()
 
@@ -659,7 +780,7 @@ class ZNP:
             counts[OneShotResponseListener],
         )
 
-    def frame_received(self, frame: GeneralFrame) -> bool:
+    def frame_received(self, frame: GeneralFrame) -> bool | None:
         """
         Called when a frame has been received. Returns whether or not the frame was
         handled by any listener.
@@ -669,7 +790,7 @@ class ZNP:
 
         if frame.header not in c.COMMANDS_BY_ID:
             LOGGER.error("Received an unknown frame: %s", frame)
-            return
+            return False
 
         command_cls = c.COMMANDS_BY_ID[frame.header]
 
@@ -680,7 +801,7 @@ class ZNP:
             # https://github.com/home-assistant/core/issues/50005
             if command_cls == c.ZDO.ParentAnnceRsp.Callback:
                 LOGGER.warning("Failed to parse broken %s as %s", frame, command_cls)
-                return
+                return False
 
             raise
 
@@ -760,7 +881,21 @@ class ZNP:
 
         return self.callback_for_responses([response], callback)
 
-    def wait_for_responses(self, responses, *, context=False) -> asyncio.Future:
+    @typing.overload
+    def wait_for_responses(
+        self, responses, *, context: typing_extensions.Literal[False] = ...
+    ) -> asyncio.Future:
+        ...
+
+    @typing.overload
+    def wait_for_responses(
+        self, responses, *, context: typing_extensions.Literal[True]
+    ) -> tuple[asyncio.Future, OneShotResponseListener]:
+        ...
+
+    def wait_for_responses(
+        self, responses, *, context: bool = False
+    ) -> asyncio.Future | tuple[asyncio.Future, OneShotResponseListener]:
         """
         Creates a one-shot listener that matches any *one* of the given responses.
         """
@@ -787,7 +922,9 @@ class ZNP:
 
         return self.wait_for_responses([response])
 
-    async def request(self, request: t.CommandBase, **response_params) -> t.CommandBase:
+    async def request(
+        self, request: t.CommandBase, **response_params
+    ) -> t.CommandBase | None:
         """
         Sends a SREQ/AREQ request and returns its SRSP (only for SREQ), failing if any
         of the SRSP's parameters don't match `response_params`.
@@ -827,7 +964,7 @@ class ZNP:
             if not request.Rsp:
                 LOGGER.debug("Request has no response, not waiting for one.")
                 self._uart.send(frame)
-                return
+                return None
 
             # We need to create the response listener before we send the request
             response_future = self.wait_for_responses(
